import ballerina/sql;
import ballerina/io;
import ballerina/time;

struct ResultPrimitive {
    int INT_TYPE;
    int LONG_TYPE;
    float FLOAT_TYPE;
    float DOUBLE_TYPE;
    boolean BOOLEAN_TYPE;
    string STRING_TYPE;
}

struct ResultSetTestAlias {
    int INT_TYPE;
    int LONG_TYPE;
    float FLOAT_TYPE;
    float DOUBLE_TYPE;
    boolean BOOLEAN_TYPE;
    string STRING_TYPE;
    int DT2INT_TYPE;
}

struct ResultObject {
    blob BLOB_TYPE;
    string CLOB_TYPE;
    blob BINARY_TYPE;
}

struct ResultMap {
    int[] INT_ARRAY;
    int[] LONG_ARRAY;
    float[] FLOAT_ARRAY;
    boolean[] BOOLEAN_ARRAY;
    string[] STRING_ARRAY;
}

struct ResultBlob {
    blob BLOB_TYPE;
}

struct ResultDates {
    string DATE_TYPE;
    string TIME_TYPE;
    string TIMESTAMP_TYPE;
    string DATETIME_TYPE;
}

struct ResultDatesStruct {
    time:Time DATE_TYPE;
    time:Time TIME_TYPE;
    time:Time TIMESTAMP_TYPE;
    time:Time DATETIME_TYPE;
}

struct ResultDatesInt {
    int DATE_TYPE;
    int TIME_TYPE;
    int TIMESTAMP_TYPE;
    int DATETIME_TYPE;
}

struct ResultSetFloat {
    float FLOAT_TYPE;
    float DOUBLE_TYPE;
    float NUMERIC_TYPE;
    float DECIMAL_TYPE;
}

struct ResultPrimitiveInt {
    int INT_TYPE;
}

struct ResultCount {
    int COUNTVAL;
}

struct ResultTest {
    int t1Row;
    int t1Int;
    int t2Row;
    int t2Int;
}

struct ResultSignedInt {
    int ID;
    int TINYINTDATA;
    int SMALLINTDATA;
    int INTDATA;
    int BIGINTDATA;
}

struct ResultComplexTypes {
    int ROW_ID;
    blob BLOB_TYPE;
    string CLOB_TYPE;
    blob BINARY_TYPE;
}

struct TestTypeData {
    int i;
    int[] iA;
    int l;
    int[] lA;
    float f;
    float[] fA;
    float d;
    boolean b;
    string s;
    float[] dA;
    boolean[] bA;
    string[] sA;
}

function testToJson () returns (json | null) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    sql:Parameter[] parameters = [];
    try {
        table dt =? testDB -> select("SELECT int_type, long_type, float_type, double_type,
                  boolean_type, string_type from DataTable WHERE row_id = 1", parameters, null);

        var result =? <json>dt;
        return result;
    } finally {
        _ = testDB -> close();
    }
    return null;
}

function testToXml () returns (xml | null) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    try {
        table dt =? testDB -> select("SELECT int_type, long_type, float_type, double_type,
                   boolean_type, string_type from DataTable WHERE row_id = 1", null, null);
        var result =? <xml>dt;
        return result;
    } finally {
        _ = testDB -> close();
    }
    return null;
}

function testToXmlMultipleConsume () returns (xml | null) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    try {
        table dt =? testDB -> select("SELECT int_type, long_type, float_type, double_type,
        boolean_type, string_type from DataTable WHERE row_id = 1", null, null);
        var result =? <xml>dt;
        io:println(result);
        return result;
    } finally {
        _ = testDB -> close();
    }
    return null;
}

function testToXmlWithAdd () returns (xml | null) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:2}
    };

    try {
        table dt1 =? testDB -> select("SELECT int_type from DataTable WHERE row_id = 1", null, null);
        var result1 =? <xml>dt1;

        table dt2 =? testDB -> select("SELECT int_type from DataTable WHERE row_id = 1", null, null);
        var result2 =? <xml>dt2;

        xml result = result1 + result2;

        table dt3 =? testDB -> select("SELECT int_type from DataTable WHERE row_id = 1", null, null);
        return result;
    } finally {
        _ = testDB -> close();
    }
    return null;
}

function testToJsonMultipleConsume () returns (json | null) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    try {
        table dt =? testDB -> select("SELECT int_type, long_type, float_type, double_type,
        boolean_type, string_type from DataTable WHERE row_id = 1", null, null);
        var result =? <json>dt;
        io:println(result);
        return result;
    } finally {
        _ = testDB -> close();
    }
    return null;
}


function toXmlComplex () returns (xml | null) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    try {
        table dt =? testDB -> select("SELECT int_type, int_array, long_type, long_array, float_type,
                    float_array, double_type, boolean_type, string_type, double_array, boolean_array, string_array
                    from MixTypes where row_id =1", null, null);
        var result =? <xml>dt;
        return result;
    } finally {
        _ = testDB -> close();
    }
    return null;
}

function testToXmlComplexWithStructDef () returns (xml | null) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    try {
        table dt =? testDB -> select("SELECT int_type, int_array, long_type, long_array, float_type,
                    float_array, double_type, boolean_type, string_type, double_array, boolean_array, string_array
                    from MixTypes where row_id =1", null, typeof TestTypeData);
        var result =? <xml>dt;
        return result;
    } finally {
        _ = testDB -> close();
    }
    return null;
}


function testToJsonComplex () returns (json | null) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    try {
        table dt =? testDB -> select("SELECT int_type, int_array, long_type, long_array, float_type,
                    float_array, double_type, boolean_type, string_type, double_array, boolean_array, string_array
                    from MixTypes where row_id =1", null, null);
        var result =? <json>dt;
        return result;
    } finally {
        _ = testDB -> close();
    }
    return null;
}


function testToJsonComplexWithStructDef () returns (json | null) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    try {
        table dt =? testDB -> select("SELECT int_type, int_array, long_type, long_array, float_type,
                    float_array, double_type, boolean_type, string_type, double_array, boolean_array, string_array
                    from MixTypes where row_id =1", null, typeof TestTypeData);
        var result =? <json>dt;
        return result;
    } finally {
        _ = testDB -> close();
    }
    return null;
}

function testJsonWithNull () returns (json | null) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    try {
        table dt =? testDB -> select("SELECT int_type, long_type, float_type, double_type,
                  boolean_type, string_type from DataTable WHERE row_id = 2", null, null);
        var result =? <json>dt;
        return result;
    }  finally {
        _ = testDB -> close();
    }
    return null;
}

function testXmlWithNull () returns (xml | null) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    try {
        table dt =? testDB -> select("SELECT int_type, long_type, float_type, double_type,
                   boolean_type, string_type from DataTable WHERE row_id = 2", null, null);
        var result =? <xml>dt;
        return result;
    } finally {
        _ = testDB -> close();
    }
    return null;
}

function testToXmlWithinTransaction () returns (string, int) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    int returnValue = 0;
    string resultXml;
    try {
        transaction {
            table dt =? testDB -> select("SELECT int_type, long_type from DataTable WHERE row_id = 1", null, null);
            var result =? <xml>dt;
            resultXml = io:sprintf("%l", [result]);
        }
        return (resultXml, returnValue);
    } finally {
        _ = testDB -> close();
    }
    return ("<fail></fail>", -1);
}

function testToJsonWithinTransaction () returns (string, int) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    int returnValue = 0;
    string result;
    try {
        transaction {
            table dt =? testDB -> select("SELECT int_type, long_type from DataTable WHERE row_id = 1", null, null);
            var jsonResult =? <json>dt;
            result = jsonResult.toString();
        }
        return (result, returnValue);
    } finally {
        _ = testDB -> close();
    }
    return ("", -2);
}

function testGetPrimitiveTypes () returns (int, int, float, float , boolean, string) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select("SELECT int_type, long_type, float_type, double_type,
              boolean_type, string_type from DataTable WHERE row_id = 1", null, typeof ResultPrimitive);
    int i;
    int l;
    float f;
    float d;
    boolean b;
    string s;
    while (dt.hasNext()) {
        var rs =? <ResultPrimitive>dt.getNext();
        i = rs.INT_TYPE;
        l = rs.LONG_TYPE;
        f = rs.FLOAT_TYPE;
        d = rs.DOUBLE_TYPE;
        b = rs.BOOLEAN_TYPE;
        s = rs.STRING_TYPE;
    }
    _ = testDB -> close();
    return (i, l, f, d, b, s);
}

function testGetComplexTypes () returns (string, string, string) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select("SELECT blob_type,clob_type,binary_type from ComplexTypes where row_id = 1",null,
                         typeof ResultObject);
    string blobValue;
    string clob;
    string binary;
    while (dt.hasNext()) {
        var rs =? <ResultObject>dt.getNext();
        blob blobData = rs.BLOB_TYPE;
        blobValue = blobData.toString("UTF-8");
        clob = rs.CLOB_TYPE;
        blob binaryData = rs.BINARY_TYPE;
        binary = binaryData.toString("UTF-8");
    }
    _ = testDB -> close();
    return (blobValue, clob, binary);
}

function testArrayData () returns (int[], int[], float[], string[], boolean[]) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select("SELECT int_array, long_array, float_array, boolean_array,
              string_array from ArrayTypes where row_id = 1", null, typeof ResultMap);
    int[] int_arr;
    int[] long_arr;
    float[] float_arr;
    string[] string_arr;
    boolean[] boolean_arr;

    while (dt.hasNext()) {
        var rs =? <ResultMap>dt.getNext();
        int_arr = rs.INT_ARRAY;
        long_arr = rs.LONG_ARRAY;
        float_arr = rs.FLOAT_ARRAY;
        boolean_arr = rs.BOOLEAN_ARRAY;
        string_arr = rs.STRING_ARRAY;
    }
    _ = testDB -> close();
    return (int_arr, long_arr, float_arr, string_arr, boolean_arr);
}

function testArrayDataInsertAndPrint () returns (int, int, int, int, int, int) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    int[] dataint = [1, 2, 3];
    float[] datafloat = [33.4, 55.4];
    string[] datastring = ["hello", "world"];
    boolean[] databoolean = [true, false, false, true, true];

    sql:Parameter paraID = {sqlType:sql:Type.INTEGER, value:4};
    sql:Parameter paraInt = {sqlType:sql:Type.ARRAY, value:dataint};
    sql:Parameter paraLong = {sqlType:sql:Type.ARRAY, value:dataint};
    sql:Parameter paraFloat = {sqlType:sql:Type.ARRAY, value:datafloat};
    sql:Parameter paraString = {sqlType:sql:Type.ARRAY, value:datastring};
    sql:Parameter paraBool = {sqlType:sql:Type.ARRAY, value:databoolean};
    sql:Parameter[] parameters = [paraID, paraInt, paraLong, paraFloat, paraString, paraBool];


   int intArrLen;
   int longArrLen;
   int floatArrLen;
   int boolArrLen;
   int strArrLen;

    int updateRet =? testDB -> update("insert into ArrayTypes(row_id, int_array, long_array, float_array,
                                string_array, boolean_array) values (?,?,?,?,?,?)", parameters);
    table dt =? testDB -> select("SELECT int_array, long_array, float_array, boolean_array, string_array
                                 from ArrayTypes where row_id = 4", null, typeof ResultMap);
    while (dt.hasNext()) {
        var rs =? <ResultMap>dt.getNext();
        io:println(rs.INT_ARRAY);
        intArrLen = lengthof rs.INT_ARRAY;
        io:println(rs.LONG_ARRAY);
        longArrLen = lengthof rs.LONG_ARRAY;
        io:println(rs.FLOAT_ARRAY);
        floatArrLen = lengthof rs.FLOAT_ARRAY;
        io:println(rs.BOOLEAN_ARRAY);
        boolArrLen = lengthof rs.BOOLEAN_ARRAY;
        io:println(rs.STRING_ARRAY);
        strArrLen = lengthof rs.STRING_ARRAY;
    }
    _ = testDB -> close();
    return (updateRet, intArrLen, longArrLen, floatArrLen, boolArrLen, strArrLen);
}

function testDateTime (int datein, int timein, int timestampin) returns (string, string, string, string) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };
    string date;
    string time;
    string timestamp;
    string datetime;

    sql:Parameter para0 = {sqlType:sql:Type.INTEGER, value:1};
    sql:Parameter para1 = {sqlType:sql:Type.DATE, value:datein};
    sql:Parameter para2 = {sqlType:sql:Type.TIME, value:timein};
    sql:Parameter para3 = {sqlType:sql:Type.TIMESTAMP, value:timestampin};
    sql:Parameter para4 = {sqlType:sql:Type.DATETIME, value:timestampin};
    sql:Parameter[] parameters = [para0, para1, para2, para3, para4];

    int count =? testDB -> update("Insert into DateTimeTypes
        (row_id, date_type, time_type, timestamp_type, datetime_type) values (?,?,?,?,?)", parameters);
    table dt =? testDB -> select("SELECT date_type, time_type, timestamp_type, datetime_type
                from DateTimeTypes where row_id = 1", null, typeof ResultDates);
    while (dt.hasNext()) {
        var rs =? <ResultDates>dt.getNext();
        time = rs.TIME_TYPE;
        date = rs.DATE_TYPE;
        timestamp = rs.TIMESTAMP_TYPE;
        datetime = rs.DATETIME_TYPE;
    }
    _ = testDB -> close();
    return (date, time, timestamp, datetime);
}

function testDateTimeAsTimeStruct () returns (int, int, int, int, int, int, int, int ) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    int dateInserted;
    int dateRetrieved;
    int timeInserted;
    int timeRetrieved;
    int timestampInserted;
    int timestampRetrieved;
    int datetimeInserted;
    int datetimeRetrieved;

    time:Time dateStruct = time:createTime(2017, 5, 23, 0, 0, 0, 0, "");
    time:Timezone zoneValue = {zoneId:"UTC"};
    time:Time timeStruct = {time:51323000, zone:zoneValue};
    time:Time timestampStruct = time:createTime(2017, 1, 25, 16, 12, 23, 0, "UTC");
    time:Time datetimeStruct = time:createTime(2017, 1, 31, 16, 12, 23, 332, "UTC");
    dateInserted = dateStruct.time;
    timeInserted = timeStruct.time;
    timestampInserted = timestampStruct.time;
    datetimeInserted = datetimeStruct.time;

    sql:Parameter para0 = {sqlType:sql:Type.INTEGER, value:31};
    sql:Parameter para1 = {sqlType:sql:Type.DATE, value:dateStruct};
    sql:Parameter para2 = {sqlType:sql:Type.TIME, value:timeStruct};
    sql:Parameter para3 = {sqlType:sql:Type.TIMESTAMP, value:timestampStruct};
    sql:Parameter para4 = {sqlType:sql:Type.DATETIME, value:datetimeStruct};
    sql:Parameter[] parameters = [para0, para1, para2, para3, para4];

    int count =? testDB -> update("Insert into DateTimeTypes
        (row_id, date_type, time_type, timestamp_type, datetime_type) values (?,?,?,?,?)", parameters);

    table dt =? testDB -> select("SELECT date_type, time_type, timestamp_type, datetime_type
                from DateTimeTypes where row_id = 31", null, typeof ResultDatesStruct);
    while (dt.hasNext()) {
        var rs =? <ResultDatesStruct>dt.getNext();
        dateRetrieved = rs.DATE_TYPE.time;
        timeRetrieved = rs.TIME_TYPE.time;
        timestampRetrieved = rs.TIMESTAMP_TYPE.time;
        datetimeRetrieved = rs.DATETIME_TYPE.time;
    }
     _ = testDB -> close();
    return (dateInserted, dateRetrieved, timeInserted, timeRetrieved, timestampInserted, timestampRetrieved,
        datetimeInserted, datetimeRetrieved);
}

function testDateTimeInt (int datein, int timein, int timestampin) returns (int, int, int, int) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    sql:Parameter para0 = {sqlType:sql:Type.INTEGER, value:32};
    sql:Parameter para1 = {sqlType:sql:Type.DATE, value:datein};
    sql:Parameter para2 = {sqlType:sql:Type.TIME, value:timein};
    sql:Parameter para3 = {sqlType:sql:Type.TIMESTAMP, value:timestampin};
    sql:Parameter para4 = {sqlType:sql:Type.DATETIME, value:timestampin};
    sql:Parameter[] parameters = [para0, para1, para2, para3, para4];

    int  date;
    int time;
    int timestamp;
    int datetime;

    int count =? testDB -> update("Insert into DateTimeTypes
        (row_id, date_type, time_type, timestamp_type, datetime_type) values (?,?,?,?,?)", parameters);
    table<ResultDatesInt> dt =? testDB -> select("SELECT date_type, time_type, timestamp_type, datetime_type
                from DateTimeTypes where row_id = 32", null, typeof ResultDatesInt);
    //table<ResultDatesInt> dt = dt1;

    while (dt.hasNext()) {
        var rs =? <ResultDatesInt>dt.getNext();
        time = rs.TIME_TYPE;
        date = rs.DATE_TYPE;
        timestamp = rs.TIMESTAMP_TYPE;
        datetime = rs.DATETIME_TYPE;
    }
    _ = testDB -> close();
    return (date, time, timestamp, datetime);
}

function testBlobData () returns (string) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    string blobStringData;
    table dt =? testDB -> select("SELECT blob_type from ComplexTypes where row_id = 1", null, typeof ResultBlob);
    blob blobData;
    while (dt.hasNext()) {
        var rs =? <ResultBlob>dt.getNext();
        blobData = rs.BLOB_TYPE;
    }
    blobStringData = blobData.toString("UTF-8");

    _ = testDB -> close();
    return blobStringData;
}

function testColumnAlias () returns (int, int, float, float, boolean, string, int) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select("SELECT dt1.int_type, dt1.long_type, dt1.float_type,
           dt1.double_type,dt1.boolean_type, dt1.string_type,dt2.int_type as dt2int_type from DataTable dt1
           left join DataTableRep dt2 on dt1.row_id = dt2.row_id WHERE dt1.row_id = 1;", null,
                                 typeof ResultSetTestAlias);

    int i;
    int l;
    float f;
    float d;
    boolean b;
    string s;
    int i2;

    while (dt.hasNext()) {
        var rs =? <ResultSetTestAlias>dt.getNext();
        i = rs.INT_TYPE;
        l = rs.LONG_TYPE;
        f = rs.FLOAT_TYPE;
        d = rs.DOUBLE_TYPE;
        b = rs.BOOLEAN_TYPE;
        s = rs.STRING_TYPE;
        i2 = rs.DT2INT_TYPE;
    }
    _ = testDB -> close();
    return (i, l, f, d, b, s, i2);
}

function testBlobInsert () returns (int) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    sql:Parameter[] params = [];
    table dt =? testDB -> select("SELECT blob_type from ComplexTypes where row_id = 1", params, typeof ResultBlob);
    blob blobData;
    while (dt.hasNext()) {
        var rs =? <ResultBlob>dt.getNext();
        blobData = rs.BLOB_TYPE;
    }
    sql:Parameter para0 = {sqlType:sql:Type.INTEGER, value:10};
    sql:Parameter para1 = {sqlType:sql:Type.BLOB, value:blobData};
    params = [para0, para1];
    int insertCount  =? testDB -> update("Insert into ComplexTypes (row_id, blob_type) values (?,?)", params);
    _ = testDB -> close();
    return insertCount;
}


function testTableAutoClose () returns (int, string) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select("SELECT int_type from DataTable WHERE row_id = 1", null, typeof ResultPrimitiveInt);
    int i;
    string test;
    while (dt.hasNext()) {
        var rs =? <ResultPrimitiveInt>dt.getNext();
        i = rs.INT_TYPE;
    }

    table dt2 =? testDB -> select("SELECT int_type, long_type, float_type, double_type,
              boolean_type, string_type from DataTable WHERE row_id = 1", null, null);
    var jsonstring =? <json> dt2;
    test = jsonstring.toString();

    _ = testDB -> select("SELECT int_type, long_type, float_type, double_type,
              boolean_type, string_type from DataTable WHERE row_id = 1", null, null);

    _ = testDB -> close();
    return (i, test);
}

function testTableManualClose () returns (int) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select("SELECT int_type from DataTable", null, typeof ResultPrimitiveInt);
    int i = 0;
    while (dt.hasNext()) {
        var rs =? <ResultPrimitiveInt>dt.getNext();
        int ret = rs.INT_TYPE;
        i = i + 1;
        if (i == 1) {
            break;
        }
    }
    dt.close();

    int data;
    table dt2 =? testDB -> select("SELECT int_type from DataTable WHERE row_id = 1", null, typeof ResultPrimitiveInt);
    while (dt2.hasNext()) {
        var rs2 =? <ResultPrimitiveInt>dt2.getNext();
        data = rs2.INT_TYPE;
    }
    dt2.close();
    _ = testDB -> close();
    return data;
}

function testCloseConnectionPool () returns (int) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select ("SELECT COUNT(*) as countVal FROM INFORMATION_SCHEMA.SYSTEM_SESSIONS", null,
                                  typeof ResultCount);
    int count;
    while (dt.hasNext()) {
        var rs =? <ResultCount> dt.getNext();
        count = rs.COUNTVAL;
    }
    _ = testDB -> close();
    return count;
}

function testTablePrintAndPrintln() {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select("SELECT int_type, long_type, float_type, double_type,
        boolean_type, string_type from DataTable WHERE row_id = 1", null, null);
    io:println(dt);
    io:print(dt);
    _ = testDB -> close();
}

function testMutltipleRows () returns (int, int) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select("SELECT int_type from DataTableRep", null, typeof ResultPrimitiveInt);

    ResultPrimitiveInt rs1 = {INT_TYPE:-1};
    ResultPrimitiveInt rs2 = {INT_TYPE:-1};
    int i = 0;
    while (dt.hasNext()) {
        if (i == 0) {
            rs1 =? <ResultPrimitiveInt>dt.getNext();
        } else {
            rs2 =? <ResultPrimitiveInt>dt.getNext();
        }
        i = i + 1;
    }
    _ = testDB -> close();
    return (rs1.INT_TYPE, rs2.INT_TYPE);
}

function testMutltipleRowsWithoutLoop () returns (int, int, int, int, string, string) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    //Iterate the whole result
    table dt =? testDB -> select("SELECT int_type from DataTableRep order by int_type desc", null,
                                 typeof ResultPrimitiveInt);
    int i1;
    int i2;
    int i3;
    int i4;
    string st1;
    string st2;
    while (dt.hasNext()) {
        var rs =? <ResultPrimitiveInt>dt.getNext();
        i1 = rs.INT_TYPE;
    }

    //Pick the first row only
    dt =? testDB -> select("SELECT int_type from DataTableRep order by int_type desc", null, typeof ResultPrimitiveInt);
    if (dt.hasNext()) {
        var rs =? <ResultPrimitiveInt>dt.getNext();
        i2 = rs.INT_TYPE;
    }
    dt.close();

    //Pick all the rows without checking
    dt =? testDB -> select("SELECT int_type from DataTableRep order by int_type desc", null, typeof ResultPrimitiveInt);
    var rs1 =? <ResultPrimitiveInt>dt.getNext();
    i3 = rs1.INT_TYPE;

    var rs2 =? <ResultPrimitiveInt>dt.getNext();
    i4 = rs2.INT_TYPE;
    dt.close();

    //Pick the first row by checking and next row without checking
    string s1 = "";
    dt =? testDB -> select("SELECT int_type from DataTableRep order by int_type desc", null, typeof ResultPrimitiveInt);
    if (dt.hasNext()) {
        var rs =? <ResultPrimitiveInt>dt.getNext();
        int i = rs.INT_TYPE;
        s1 = s1 + i;
    }

    var rs =? <ResultPrimitiveInt>dt.getNext();
    int i = rs.INT_TYPE;
    s1 = s1 + "_" + i;

    if (dt.hasNext()) {
        s1 = s1 + "_" + "HAS";
    } else {
        s1 = s1 + "_" + "NOT";
    }

    //Pick the first row without checking, then check and no fetch, and finally fetch row by checking
    string s2 = "";
    dt =? testDB -> select("SELECT int_type from DataTableRep order by int_type desc", null, typeof ResultPrimitiveInt);
    rs =? <ResultPrimitiveInt>dt.getNext();
    i = rs.INT_TYPE;
    s2 = s2 + i;
    if (dt.hasNext()) {
        s2 = s2 + "_" + "HAS";
    } else {
        s2 = s2 + "_" + "NO";
    }
    if (dt.hasNext()) {
        s2 = s2 + "_" + "HAS";
    } else {
        s2 = s2 + "_" + "NO";
    }
    if (dt.hasNext()) {
        rs =? <ResultPrimitiveInt>dt.getNext();
        i = rs.INT_TYPE;
        s2 = s2 + "_" + i;
    }
    if (dt.hasNext()) {
        s2 = s2 + "_" + "HAS";
    } else {
        s2 = s2 + "_" + "NO";
    }
    if (dt.hasNext()) {
        s2 = s2 + "_" + "HAS";
    } else {
        s2 = s2 + "_" + "NO";
    }
    _ = testDB -> close();
    return (i1, i2, i3, i4, s1, s2);
}

function testHasNextWithoutConsume () returns (boolean, boolean, boolean) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select("SELECT int_type from DataTableRep order by int_type desc", null,
                                 typeof ResultPrimitiveInt);
    boolean b1 = false;
    boolean b2 = false;
    boolean b3 = false;

    if (dt.hasNext()) {
        b1 = true;
    }
    if (dt.hasNext()) {
        b2 = true;
    }
    if (dt.hasNext()) {
        b3 = true;
    }
    _ = testDB -> close();
    return (b1, b2, b3);
}

function testGetFloatTypes () returns (float, float, float, float) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select("SELECT float_type, double_type,
                  numeric_type, decimal_type from FloatTable WHERE row_id = 1", null, typeof ResultSetFloat);
    float f;
    float d;
    float num;
    float dec;

    while (dt.hasNext()) {
        var rs =? <ResultSetFloat>dt.getNext();
        f = rs.FLOAT_TYPE;
        d = rs.DOUBLE_TYPE;
        num = rs.NUMERIC_TYPE;
        dec = rs.DECIMAL_TYPE;
    }
    _ = testDB -> close();
    return (f, d, num, dec);
}

function testSignedIntMaxMinValues () returns (int, int, int, string, string, string) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    string insertSQL = "INSERT INTO IntegerTypes(id,tinyIntData, smallIntData, intData, bigIntData) VALUES (?,?, ?,?,?)";
    string selectSQL = "SELECT id,tinyIntData,smallIntData,intData,bigIntData FROM IntegerTypes";

    int maxInsert;
    int minInsert;
    int nullInsert;
    string jsonStr;
    string xmlStr;
    string str;

    //Insert signed max
    sql:Parameter para1 = {sqlType:sql:Type.INTEGER, value:1};
    sql:Parameter para2 = {sqlType:sql:Type.TINYINT, value:127};
    sql:Parameter para3 = {sqlType:sql:Type.SMALLINT, value:32767};
    sql:Parameter para4 = {sqlType:sql:Type.INTEGER, value:2147483647};
    sql:Parameter para5 = {sqlType:sql:Type.BIGINT, value:9223372036854775807};
    sql:Parameter[] parameters = [para1, para2, para3, para4, para5];
    maxInsert =? testDB -> update(insertSQL, parameters);

    //Insert signed min
    para1 = {sqlType:sql:Type.INTEGER, value:2};
    para2 = {sqlType:sql:Type.TINYINT, value:-128};
    para3 = {sqlType:sql:Type.SMALLINT, value:-32768};
    para4 = {sqlType:sql:Type.INTEGER, value:-2147483648};
    para5 = {sqlType:sql:Type.BIGINT, value:-9223372036854775808};
    parameters = [para1, para2, para3, para4, para5];
    minInsert =?  testDB -> update(insertSQL, parameters);

    //Insert null
    para1 = {sqlType:sql:Type.INTEGER, value:3};
    para2 = {sqlType:sql:Type.TINYINT, value:null};
    para3 = {sqlType:sql:Type.SMALLINT, value:null};
    para4 = {sqlType:sql:Type.INTEGER, value:null};
    para5 = {sqlType:sql:Type.BIGINT, value:null};
    parameters = [para1, para2, para3, para4, para5];
    nullInsert =? testDB -> update(insertSQL, parameters);

    table dt =? testDB -> select(selectSQL, null, null);
    var j =? <json>dt;
    jsonStr = j.toString();

    dt =? testDB -> select(selectSQL, null, null);
    var x =? <xml>dt;
    xmlStr = io:sprintf("%l", [x]);

    dt =? testDB -> select(selectSQL, null, typeof ResultSignedInt);
    str = "";
    while (dt.hasNext()) {
        var result =? <ResultSignedInt>dt.getNext();
        str = str + result.ID + "|" + result.TINYINTDATA + "|" + result.SMALLINTDATA + "|" + result.INTDATA + "|" +
              result.BIGINTDATA + "#";
    }
    _ = testDB -> close();
    return (maxInsert, minInsert, nullInsert, jsonStr, xmlStr, str);
}

function testComplexTypeInsertAndRetrieval () returns (int, int, string, string, string) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    string insertSQL = "INSERT INTO ComplexTypes(row_id, blob_type, clob_type, binary_type) VALUES (?,?,?,?)";
    string selectSQL = "SELECT row_id, blob_type, clob_type, binary_type FROM ComplexTypes where row_id = 100 or row_id = 200";
    string text = "Sample Text";
    blob content = text.toBlob("UTF-8");

    int retDataInsert;
    int retNullInsert;
    string jsonStr;
    string xmlStr;
    string str;

    //Insert data
    sql:Parameter para1 = {sqlType:sql:Type.INTEGER, value:100};
    sql:Parameter para2 = {sqlType:sql:Type.BLOB, value:content};
    sql:Parameter para3 = {sqlType:sql:Type.CLOB, value:text};
    sql:Parameter para4 = {sqlType:sql:Type.BINARY, value:content};
    sql:Parameter[] parameters = [para1, para2, para3, para4];
    retDataInsert =? testDB -> update(insertSQL, parameters);

    //Insert null values
    para1 = {sqlType:sql:Type.INTEGER, value:200};
    para2 = {sqlType:sql:Type.BLOB, value:null};
    para3 = {sqlType:sql:Type.CLOB, value:null};
    para4 = {sqlType:sql:Type.BINARY, value:null};
    parameters = [para1, para2, para3, para4];
    retNullInsert =? testDB -> update(insertSQL, parameters);

    table dt =? testDB -> select(selectSQL, null, null);
    var j =? <json>dt;
    jsonStr = j.toString();

    dt =? testDB -> select(selectSQL, null, null);
    var x =? <xml>dt;
    xmlStr = io:sprintf("%l", [x]);

    dt =? testDB -> select(selectSQL, null, typeof ResultComplexTypes);
    str = "";
    while (dt.hasNext()) {
        var result =? <ResultComplexTypes>dt.getNext();
        str = str + result.ROW_ID + "|" + result.BLOB_TYPE.toString("UTF-8") + "|" + result.CLOB_TYPE + "|";
    }
    _ = testDB -> close();
    return (retDataInsert, retNullInsert, jsonStr, xmlStr, str);
}

function testJsonXMLConversionwithDuplicateColumnNames () returns (string, string) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select("SELECT dt1.row_id, dt1.int_type, dt2.row_id, dt2.int_type from DataTable dt1 left
            join DataTableRep dt2 on dt1.row_id = dt2.row_id WHERE dt1.row_id = 1", null, null);
    var j =? <json> dt;
    string jsonStr = j.toString();

    table dt2 =? testDB -> select("SELECT dt1.row_id, dt1.int_type, dt2.row_id, dt2.int_type from DataTable dt1 left
            join DataTableRep dt2 on dt1.row_id = dt2.row_id WHERE dt1.row_id = 1", null, null);
    var x =? <xml> dt2;
    //string xmlStr = <string> x; //TODO
    string xmlStr;

    _ = testDB -> close();
    return (jsonStr, xmlStr);
}

function testStructFieldNotMatchingColumnName () returns (int, int, int, int, int) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select("SELECT count(*) from DataTable WHERE row_id = 1", null, typeof ResultCount);
    int countAll;
    int i1;
    int i2;
    int i3;
    int i4;
    while (dt.hasNext()) {
        var rs =? <ResultCount>dt.getNext();
        countAll = rs.COUNTVAL;
    }

    table dt2 =? testDB -> select("SELECT dt1.row_id, dt1.int_type, dt2.row_id, dt2.int_type from DataTable dt1 left
            join DataTableRep dt2 on dt1.row_id = dt2.row_id WHERE dt1.row_id = 1", null, typeof ResultTest);
    while (dt2.hasNext()) {
        var rs =? <ResultTest>dt2.getNext();
        i1 = rs.t1Row;
        i2 = rs.t1Int;
        i3 = rs.t2Row;
        i4 = rs.t2Int;
    }
    _ = testDB -> close();
    return (countAll, i1, i2, i3, i4);
}

function testGetPrimitiveTypesWithForEach () returns (int, int, float, float, boolean, string) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt1 =? testDB -> select("SELECT int_type, long_type, float_type, double_type,
              boolean_type, string_type from DataTable WHERE row_id = 1", null, typeof ResultPrimitive);

    table<ResultPrimitive> dt = dt1; //TODO:Remove this
    int i;
    int l;
    float f;
    float d;
    boolean b;
    string s;
    foreach x in dt {
        i = x.INT_TYPE;
        l = x.LONG_TYPE;
        f = x.FLOAT_TYPE;
        d = x.DOUBLE_TYPE;
        b = x.BOOLEAN_TYPE;
        s = x.STRING_TYPE;
    }
    _ = testDB -> close();
    return (i, l, f, d, b, s);
}

function testMutltipleRowsWithForEach () returns (int, int) {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt1 =? testDB -> select("SELECT int_type from DataTableRep", null, typeof ResultPrimitiveInt);
    table<ResultPrimitiveInt> dt = dt1; //TODO:Remove this

    ResultPrimitiveInt rs1 = {INT_TYPE: -1};
    ResultPrimitiveInt rs2 = {INT_TYPE: -1};
    int i = 0;
    foreach x in dt {
        if (i == 0) {
            rs1 = x;
        } else {
            rs2 = x;
        }
        i = i + 1;
    }
    _ = testDB -> close();
    return (rs1.INT_TYPE, rs2.INT_TYPE);
}

function testTableAddInvalid () {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select("SELECT int_type from DataTableRep", null, typeof ResultPrimitiveInt);
    try {
        ResultPrimitiveInt row = {INT_TYPE:443};
        dt.add(row);
    } finally {
        _ = testDB -> close();
    }
}

function testTableRemoveInvalid () {
    endpoint sql:Client testDB {
        database: sql:DB.HSQLDB_FILE,
        host: "./target/tempdb/",
        port: 0,
        name: "TEST_DATA_TABLE_DB",
        username: "SA",
        password: "",
        options: {maximumPoolSize:1}
    };

    table dt =? testDB -> select("SELECT int_type from DataTableRep", null, typeof ResultPrimitiveInt);
    try {
        ResultPrimitiveInt row = {INT_TYPE:443};
        _ = dt.remove(isDelete);
    } finally {
       _ = testDB -> close();
    }
}

function isDelete (ResultPrimitiveInt p) returns (boolean) {
    return p.INT_TYPE < 2000;
}
